/*
much: a program to produce mutliple-choice tests
Copyright (C) 2004, 2006  Mihail N. Kolountzakis

Read the file COPYRIGHT for the complete copyright.
*/

#include	"vars.h"

// The following macro prepends the basedir path to a filename
static	char	__REL__[10000];
#define	REL(s)	(sprintf(__REL__,"%s/%s", basedir, s),__REL__)

// generate a random subset of k elements of the set {0, ..., N-1} and
// return the elements in the vector elem
void    random_set(int N, int k, int *elem)
{
        int     n=0, i, r;
        char    *s;
        double  x;

        s = (char*)malloc(N);
        for(i=0; i<N; i++) s[i]=0;

        while(n<k) {
	 x = RandomUniform();
         // x = drand48();
         x *= N;
         r = (int)x;
         if(s[r]) continue;
         s[r] = 1;
         elem[n++] = r;
        }
	free(s);
}

// make a tex file with all questions to aid the instructor to answer
// to have an overview of the tests to be produced
int	output_all_questions()
{
	char	qname[MAX_STR_LENGTH];
	FILE	*f;
	int	i, j, k, nfields;

	f = fopen(REL(all_fname), "w");
	if(!f) {
	 fprintf(llog, "*** output_all_questions: Cannot open '%s'. Quitting.\n", all_fname);
	 return 1;
	}
	// loop over all question types
	for(i=0; i<num_types; i++) {
	 // loop over all questions of the particular type
	 for(j=0; j<existing_num_per_type[i]; j++) {
	  strcpy(qname, question_fname[i][j]);
	  nfields = read_question_file(qname);
//	  fprintf(f, "\n\n\\noindent\n{\\tt [\\tl{%s}]} %s\n\n", qname, line[0]);
	  // what was this \tl for, above? (20060305)
	  fprintf(f, "\n\n\\noindent\n{\\tt [{%s}]} %s\n\n", qname, line[0]);
	  for(k=0; k<nfields-1; k++)
	   fprintf(f, "{%s}: %s\\ \\ \n", ans_name[k], line[1+k]);
	 }
	}
	fclose(f);

}

// read the question file into memory, return how many text fields, 0 on error
// each text field in the question file must terminate with a line starting
// with a period (.), even the last text field
// the fields read are stored in the global variable line[][]
int	read_question_file(char *fname)
{
	char	text_line[MAX_STR_LENGTH], text_multiline[10*MAX_STR_LENGTH]="";
	int	fields_read=0;
	FILE	*f;

	f = fopen(REL(fname), "r");
	if(!f) {
	 fprintf(stderr, "Cannot open '%s'\n", fname);
	 return 0;
	}
read_line:
	fgets(text_line, MAX_STR_LENGTH-1, f);	
	if(feof(f)) {
	 fclose(f);
	 return fields_read;
	}
	if(text_line[0] != '.') // keep reading until you find a line starting with a dot (.)
	 strcat(text_multiline, text_line);
	else {
	 // remove newline from end of string; internal newlines are kept
	 if(text_multiline[strlen(text_multiline)-1]=='\n')
          text_multiline[strlen(text_multiline)-1]=0;
	 strcpy(line[fields_read], text_multiline);
	 text_multiline[0]=0;
	 fields_read++;
	}
	goto read_line;
}
	
// generate the tests
main_make_tests(int num_papers, int first_serial)
{
	int	i, j, k, r, s, serial, ret, num_desired, elem[1000];
	char	answers[MAX_STR_LENGTH], test_type[MAX_STR_LENGTH];
	double	pen[MAX_QUESTIONS_PER_PAPER];
	FILE	*f;

	// open output files for writing
	out = fopen(REL(output_fname), "w");
	if(!out) {
	 fprintf(llog, "%d: Cannot open '%s'. Quitting.\n", yylineno, output_fname);
	 return 1;
	}

	ser = fopen(REL(serial_fname), "w");
	if(!ser) {
	 fprintf(llog, "%d: *** Cannot open '%s'. Quitting.\n", yylineno, serial_fname);
	 return 1;
	}
	fprintf(ser, "%% random seed used: %d\n", seed);

	ann = fopen(REL(announce_fname), "w");
	if(!ann) {
	 fprintf(llog, "%d: *** Cannot open '%s'. Quitting.\n", yylineno, announce_fname);
	 return 1;
	}
	fprintf(ann, "%+20s %+20s %+40s\n\n",
	 "serial number",
	 "correct answers",
	 "penalties for wrong answers");

	fprintf(out, "%% random seed used: %d\n", seed);
	fprintf(out, "\\mcglobalheader\n");
	// loop over all papers
	for(i=0,serial=first_serial; i<num_papers; i++,serial++) {
	 static char q[MAX_QUESTION_TYPES][MAX_NAME], p[MAX_QUESTION_TYPES][MAX_NAME];
	 static int ren[MAX_QUESTION_TYPES];

	 //fprintf(llog, "%d: Generating paper No %d with S/N %d\n", yylineno, i, serial);

	 fprintf(out, "\\def\\mcserialnumber{%d}\n", serial);
	 fprintf(out, "\\mcpaperheader\n");
	 fprintf(ser, "%3d ", serial);
	 fprintf(ann, "%20d ", serial);

	 // decide which question per question type
	 for(r=j=0; j<num_types; j++) {
	  random_set(existing_num_per_type[j], num_per_type[j], elem);
	  for(k=0; k<num_per_type[j]; k++) {
	   strcpy(q[r++], question_fname[j][elem[k]]);
	  }
	 }

	 // shuffle at random the order of the questions
	 random_set(questions_per_paper, questions_per_paper, ren);
	 for(r=0; r<questions_per_paper; r++) strcpy(p[ren[r]], q[r]);
	 for(r=0; r<questions_per_paper; r++) strcpy(q[r], p[r]);

	 // record the question of the test next to the serial number
	 for(r=0; r<questions_per_paper; r++)
	  fprintf(ser, "%s ", q[r]);

	 // Now output the chosen questions in files output and serial
	 for(r=0; r<questions_per_paper; r++) {
	  int ret, perm[100], s, special;
	  ret = read_question_file(q[r]);
	  if(!ret) {
	   fprintf(llog, "*** Error reading file '%s'. Quitting.\n", q[r]);
	   return 1;
	  }
	  special = (line[0][0] == '?'); // UNDOCUMENDED FEATURE: question text starts with '?'. record '*' as a correct answer
	  pen[r] = 1./(ret-2); // what is the penalty for r-th question
	  random_set(ret-1, ret-1, perm); // perm[s] is which answer is written in s-th position
	  for(s=0; s<ret-1; s++)
	   if(!perm[s]) {
	    fprintf(ser, "%c", special?'*':letter[s]);
	    answers[r] = special?'*':letter[s];
	   }

	  // print question
	  fprintf(out, "\n\n\\def\\mcquestionnumber{%d}\n", r+1);
	  fprintf(out, "\n\n\\mcquestionheader ");
	  fprintf(out, "%s\\\\\n", line[0]);
	  for(s=0; s<ret-1; s++)
	   fprintf(out, "{%s}: %s\\ \\ ", ans_name[s], line[1+perm[s]]);
	  fprintf(out, "\n\n\\mcquestionfooter\n\n");
	 }

	 fprintf(out, "\n\n\\mcpaperfooter\n\n");
	 fprintf(ser, "\n");
	 answers[questions_per_paper]=0;
	 fprintf(ann, "%+20s ", answers);
	 for(r=0; r<questions_per_paper; r++)
	  fprintf(ann, "%5.3f ", pen[r]);
	 fprintf(ann, "\n");
	}

	// close output files
	fclose(out);
	fclose(ser);
	fclose(ann);


	output_all_questions();
}

// compare two rules by serial numbers
int	compare(const void *a, const void *b)
{
	rule	*A=(rule*)a, *B=(rule*)b;
	if(A->serial < B->serial) return -1;
	else
	if(A->serial > B->serial) return 1;
	else
	return 0;
}

// grade the papers
int	make_grades(char *correct, char *tests, char *grades)
{
	FILE	*r, *t, *g;
	char	line[MAX_STR_LENGTH];
	static	rule	rr[MAX_RULES], *rp;
	char	word[MAX_STR_LENGTH];
	static	student	st[MAX_STUDENTS];
	int	i, j, k, l, lc, linesread, ret, num_serials;

	// read the correct answers
	fprintf(llog, "%d: opening file '%s' to read correct answers\n", yylineno, correct);
	fflush(llog);
	r = fopen(REL(correct), "r");
	if(!r) {
	 fprintf(llog, "%d: *** Cannot open '%s'. Quitting.\n", yylineno, correct);
	 return 1;
	}
	lc = 0;
	fgets(word, 999, r);
	fgets(word, 999, r);
	while(2 == fscanf(r, "%d %s", &i, word)) {
	 int j, stars=0, nq=strlen(word);
	 for(j=0; j<nq; j++) stars += (word[j]=='*'); // not used for the moment
	 strcpy(rr[lc].correct, word);
	 rr[lc].serial = i;
	 for(j=0; j<nq; j++)
	  if(1 != fscanf(r, "%lf ", &(rr[lc].penalty[j])))
	   fprintf(llog, "%d: *** Error reading '%s'. Quitting.\n", yylineno, correct);
	 rr[lc].rescale = 1.0; // for the moment
	 lc++;
	}
	fclose(r);
	num_serials = lc;
	qsort((void*)rr, num_serials, sizeof(rule), compare);
	fprintf(llog, "%d: read and sorted %d serial-answer pairs.\n", yylineno, num_serials);
	fflush(llog);

	// read the actual answers and grade them at the same time
	fprintf(llog, "%d: opening file '%s' to read actual answers\n", yylineno, tests);
	fflush(llog);
	t = fopen(REL(tests), "r");
	if(!t) {
	 fprintf(llog, "%d: *** cannot open '%s', quitting.\n", yylineno, tests);
	 return 1;
	}

	lc = 0;
	linesread = 0;
	while(fgets(line, 999, t)) {
	 int	serial, ret;
	 char	ans[MAX_QUESTIONS_PER_PAPER], descr[MAX_STR_LENGTH];
	 rule	rrr;
	 double	score;


	 linesread++;

	 fprintf(llog, "%d: reading line %d\n", yylineno, linesread);
	 fflush(llog);

	 ret = sscanf(line, "%d %s %200[^\n]", &serial, ans, descr); // descr may contain white space

	 // bad line
	 if(ret != 3) {
	  fprintf(llog, "%d: file '%s': ignoring line %d (%d/3 fields read), continuing\n",
	   yylineno, tests, linesread, ret);
	  fflush(llog);
	  continue;
	 }

	 // log OK line
	 fprintf(llog, "%d: serial=%d, ans=%s, descr=%s\n",
	  yylineno, serial, ans, descr);
	 fflush(llog);

	 // check that we know how to grade this serial number
	 rrr.serial = serial;
	 rp = (rule *)bsearch(&rrr, (void*)rr, num_serials, sizeof(rule), compare);
	 if(!rp) {
	  fprintf(llog, "%d: *** File '%s': error in line %d:\
 serial number %d not found in database, continuing\n", yylineno, tests, linesread, serial);
	  fflush(llog);
	  continue;
	 }

	 l = strlen(rp->correct);
	 // correct answers and actual answers differ in number; reject
	 if(strlen(ans) != l) {
	  fprintf(llog, "%d: *** File '%s': error in line %d:\
 answer string has incorrect length\n", yylineno, tests, linesread, serial);
	  continue;
	 }

	 // seems OK. grade it
	 for(score=0.0,i=0; i<l; i++) {
	  if(rp->correct[i] == '*') { score += 1.0; continue; } // UNDOCUMENDED FEATURE: question always taken correct
	  							// if correct answer is '*'
	  if(ans[i] == '-') continue; // no change in current score for no answer
	  if(ans[i] == rp->correct[i]) score += 1.0; // correct
	  else score -= rp->penalty[i]; // wrong
	 }
	 if(score < 0.) score = 0.;
	 score *= maxgrade/l; // ignore negative score, normalize to required scale

	 // grading finished. record student's paper
	 strcpy(st[lc].descr, descr);
	 strcpy(st[lc].ans, ans);
	 strcpy(st[lc].correct, rp->correct);
	 st[lc].serial = serial;
	 st[lc].grade = score;

	 lc++;
	 if(feof(t)) break;
	}

	// output
	fprintf(llog, "%d: opening file '%s' to write the grades\n", yylineno, grades);
	fflush(llog);
	g = fopen(REL(grades), "w");
	if(!g) {
	 fprintf(llog, "%d: *** Cannot open '%s'. Quitting.\n", yylineno, grades);
	 return 1;
	}
	fprintf(g, "%10s %-15s %-15s %7s %-40s\n\n",
	 "serial", "answers", "correct", "grade", "description");
	fflush(g);
	for(i=0; i<lc; i++) {
	 fprintf(g, "%10d %-15s %-15s %#7.2f %-40s\n",
	  st[i].serial, st[i].ans, st[i].correct, st[i].grade, st[i].descr);
	}

	fprintf(llog, "%d: read and processed %d actual answers.\n", yylineno, lc);
	fflush(llog);

	fclose(t);
	fclose(g);

}

main(int argc, char **argv)
{
	char	ch, line[MAX_STR_LENGTH];

	if(argc > 3) {
usage:
	 fprintf(stderr, "usage: %s (create | grade) <test description file>\n", argv[0]);
	 exit(1);
	}

	// open log file
        llog = fopen(REL(log_fname), "w");

		get_args:
	if(argc == 1) { // no arguments were given
	 printf("(c)reate or (g)rade a test? ");
	 if(!fgets(line, MAX_STR_LENGTH-1, stdin)) goto get_args;
	 ch = toupper(line[0]);
	 if((ch != 'C') && (ch != 'G')) goto get_args;
	 else mode = ch;
	 printf("test description file: ");
	 if(!fgets(line, MAX_STR_LENGTH-1, stdin)) goto get_args;
	 line[strlen(line)-1] = 0;
	 yyin = fopen(line, "r");
	 if(!yyin) {
          fprintf(llog, "cannot open %s\n", line);
          fprintf(stderr, "cannot open %s\n", line);
          exit(1);
         }
	 yyparse();
	 fclose(llog);
	 fclose(yyin);
	 exit(0);
	}


	// one or two args were given
	if(!strcmp(argv[1], "create")) {
	 mode = 'C';
	 if(argc == 3) yyin = fopen(argv[2], "r");
	 else yyin = stdin;
	 if(!yyin) {
	  fprintf(llog, "cannot open %s\n", argv[2]);
	  fprintf(stderr, "cannot open %s\n", argv[2]);
	  exit(1);
	 }
	 yyparse();
	 fclose(llog);
	 fclose(yyin);
	 exit(0);
	}
	if(!strcmp(argv[1], "grade")) {
	 mode = 'G';
	 if(argc == 3) yyin = fopen(argv[2], "r");
	 else yyin = stdin;
         if(!yyin) {
          fprintf(llog, "cannot open %s\n", argv[2]);
          fprintf(stderr, "cannot open %s\n", argv[2]);
          exit(1);
         }
	 yyparse();
	 fclose(llog);
	 fclose(yyin);
	 exit(0);
	}

	goto usage;
}
