%{

#define	MAIN_PROGRAM

#include "parse.tab.h"
#include "vars.h"

%}

%union	{ int i; double d; char *p; }

%token	<i>INTEGER
%token	<d>REAL
%token	<p>ID
%token	<p>STRING
%token	PLUS	MINUS	TIMES	DIVIDE	POWER
%token	LEFT_PARENTHESIS	RIGHT_PARENTHESIS
%token	USE SEED SERIAL CREATE GRADE FROM TO SCALE DIRECTORY

%type	<d>Number

%left	PLUS	MINUS
%left	TIMES	DIVIDE
%left	NEG
%right	POWER

%start Input
%%


Input:
	  /* Empty */
	| Input Statement
	;
Statement:
	  DIRECTORY STRING ';'
	  {
	    strcpy(basedir, $2);
	    fprintf(llog, "%d: base dir set to %s\n", yylineno, basedir);
	  }

	| SEED INTEGER ';'
	  {
	    if(mode != 'C') {
	     fprintf(llog, "%d: ignoring seed command in grading mode\n", yylineno);
	     goto seed_end; // if in grading mode do nothing here
	    }
	    seed = $2;
	    RandomInitialise(seed, seed);
	    // srand48(seed = $2);
	    fprintf(llog, "%d: seed set to %d, random number generator initialized\n", yylineno, seed);
		seed_end:;
	  }

	| USE INTEGER FROM STRING ';'
	  { 
            DIR     *dir;
            struct dirent   *entry;
            char   name[MAX_STR_LENGTH];
            int    ret, exist=0;

	    if(mode != 'C') {
	     fprintf(llog, "%d: ignoring category command in grading mode\n", yylineno);
	     goto category_end; // if in grading mode do nothing here
	    }
	    strcpy(type[num_types], $4);
	    num_per_type[num_types] = $2;
	    num_types++;
	    fprintf(llog, "%d: question category %s, required number of questions %d\n",
	     yylineno, type[num_types-1], num_per_type[num_types-1]);
	    fflush(llog);

            // find number of existing questions for this type, as well as the filenames
	    // quit if not enough
            sprintf(name, "%s", type[num_types-1]);
            if((dir = opendir(basedir)) == NULL) {
             perror("opendir");
             return 1;
            }
            while((entry = readdir(dir)) != NULL){
             ret = fnmatch(name, entry->d_name, 0);
             if(!ret){
              fprintf(llog, "%d: %s matches %s\n", yylineno, entry->d_name, name);
	      fflush(llog);
              strcpy(question_fname[num_types-1][exist], entry->d_name);
              exist++;
             } 
            }
            existing_num_per_type[num_types-1] = exist;
            if(existing_num_per_type[num_types-1] < num_per_type[num_types-1]) {
             fprintf(llog, "%d: *** Too few question files for question '%s'. Quitting.\n",
	      yylineno, type[num_types-1]);
             return 1;
            }
            fprintf(llog, "%d: %d questions of type '%s' were found\n",
	     yylineno, existing_num_per_type[num_types-1], type[num_types-1]);
            closedir(dir);
		category_end:;
	  }

	| SERIAL INTEGER ';'
	  {
	    if(mode != 'C') {
	     fprintf(llog, "%d: ignoring serial command in grading mode\n", yylineno);
	     goto serial_end; // if in grading mode do nothing here
	    }
	    first_serial = $2;
	    fprintf(llog, "%d: first serial number set to %d\n", yylineno, first_serial);
		serial_end:;
	  }

	| CREATE INTEGER ';'
	  {
	    int i, j;
	
	    if(mode != 'C') {
	     fprintf(llog, "%d: ignoring create command in grading mode\n", yylineno);
	     goto create_end; // if in grading mode do nothing here
	    }
	    questions_per_paper = 0;
	    for(i=0; i<num_types; i++) questions_per_paper += num_per_type[i];
	    // report what's read for checking
	    fprintf(llog, "%d: number of test types: %d\n", yylineno, num_types);
	    for(i=0; i<num_types; i++)
	     fprintf(llog, "%d: type %s, number desired: %3d\n", yylineno, type[i], num_per_type[i]);
	    num_papers = $2;
	    main_make_tests(num_papers, first_serial);
		create_end:;
	  }

	| SCALE Number ';'
          {
            if(mode != 'G') {
             fprintf(llog, "%d: ignoring scale command in creating mode\n", yylineno);
             goto scale_end; // if in creating mode do nothing here
            }
	    maxgrade = $2;
            fprintf(llog, "%d: set scale to %g\n", yylineno, maxgrade);
                scale_end:;
          }

        | GRADE FROM STRING TO STRING ';'
          {
            if(mode != 'G') {
             fprintf(llog, "%d: ignoring grade command in creating mode\n", yylineno);
             goto grade_end; // if in creating mode do nothing here
            }
            strcpy(papers_fname, $3);
            strcpy(grades_fname, $5);
            fprintf(llog, "%d: making grades from answers in %s to file %s\n",
             yylineno, papers_fname, grades_fname);
            make_grades(announce_fname, papers_fname, grades_fname);
                grade_end:;
          }


Number	:	INTEGER
	{ $$ = $1; }
	|	REAL
	{ $$ = $1; }
%%

int yyerror(char *s) {
  fprintf(llog, "line %d: %s\n", yylineno, s);
  fprintf(stderr, "line %d: %s\n", yylineno, s);
}
